# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for communicating using NNG, see https://pynng.readthedocs.io.

.. deprecated:: 0.2.0
   Use :mod:`cicada.communicator.socket` instead.

"""

import collections
import contextlib
import functools
import hashlib
import itertools
import logging
import math
import multiprocessing
import numbers
import os
import pickle
import queue
import threading
import time
import traceback
import warnings

import numpy
import pynng

from .interface import Communicator
import cicada
import cicada.bind

log = logging.getLogger(__name__)
logging.getLogger("pynng.nng").setLevel(logging.INFO)

Message = collections.namedtuple("Message", ["serial", "tag", "sender", "payload"])
Message.__doc__ = """
Wrapper class for messages sent between processes.
"""

def nng_timeout(value):
    """Convert a timeout in seconds to an NNG timeout."""
    return -1 if value is None else int(float(value) * 1000.0)


class Failed(Exception):
    """Used to indicate that a player process raised an exception."""
    def __init__(self, exception, traceback):
        self.exception = exception
        self.traceback = traceback

    def __repr__(self):
        return f"Failed(exception={self.exception!r})" # pragma: no cover


class Terminated(Exception):
    """Used to indicate that a player process terminated unexpectedly without output."""
    def __init__(self, exitcode):
        self.exitcode = exitcode

    def __repr__(self):
        return f"Terminated(exitcode={self.exitcode!r})" # pragma: no cover


class Revoked(Exception):
    """Raised calling an operation after the communicator has been revoked."""
    pass


class Timeout(Exception):
    """Raised when a blocking operation has timed-out."""
    pass


class TryAgain(Exception):
    """Raised when a non-blocking operation would block."""
    pass


class NNGCommunicator(Communicator):
    """Cicada communicator that uses pynng (https://pynng.readthedocs.io) as the transport layer.

    .. deprecated:: 0.2.0
       Use :class:`cicada.communicator.socket.SocketCommunicator` instead.

    Note
    ----
    Creating a communicator is a collective operation that must be called by
    all players that will be members.

    Parameters
    ----------
    name: string, optional
        The name of this communicator, which is used strictly for logging
        and debugging.  If unspecified the default is "world".
    world_size: integer, optional
        The number of players that will be members of this communicator.
        Defaults to the value of the WORLD_SIZE environment variable.
    link_addr: string, optional
        NNG address of the root (rank 0) player.  This address must be
        publically accessible to all of the other players.  Defaults to the
        value of the LINK_ADDR environment variable.
    rank: integer, optional
        The rank of the local player, in the range [0, world_size).  Defaults
        to the value of the RANK environment variable.
    host_addr: string, optional
        NNG address of the local player.  This address must be publically
        accessible to all of the other players.  Defaults to the value of the
        HOST_ADDR environment variable.  Note that this value is ignored
        by the root player.
    timeout: number or `None`
        Maximum time to wait for normal communication to complete in seconds, or `None` to disable timeouts.
    setup_timeout: number or `None`
        Maximum time allowed to setup the communicator in seconds, or `None` to disable timeouts during setup.
    """

    _tags = [
        "allgather",
        "barrier-enter",
        "barrier-exit",
        "broadcast",
        "gather",
        "gatherv",
        "revoke",
        "scatter",
        "scatterv",
        "send",
        "shrink-enter",
        "shrink-exit",
        "split",
        "split-prepare",
        ]

    class _Done(object):
        """Sentinel message used to shut-down the queueing thread."""
        pass

    def __init__(self, *, name=None, world_size=None, rank=None, link_addr=None, host_addr=None, token=0, timeout=5, setup_timeout=5):
        warnings.warn("cicada.communicator.nng.NNGCommunicator is deprecated, use cicada.communicator.socket.SocketCommunicator instead.", cicada.DeprecationWarning, stacklevel=2)
        # Setup defaults.
        if name is None:
            name = "world"
        if world_size is None:
            world_size = int(os.environ["WORLD_SIZE"])
        if rank is None:
            rank = int(os.environ["RANK"])
        if link_addr is None:
            link_addr = os.environ["LINK_ADDR"]
        if host_addr is None:
            host_addr = os.environ["HOST_ADDR"]

        # Enforce preconditions.
        if not isinstance(world_size, int):
            raise ValueError("world_size must be an integer.") # pragma: no cover
        if not world_size > 0:
            raise ValueError("world_size must be an integer greater than zero.") # pragma: no cover
        if not isinstance(rank, int):
            raise ValueError("rank must be an integer.") # pragma: no cover
        if not (0 <= rank and rank < world_size):
            raise ValueError(f"rank must be in the range [0, {world_size}).") # pragma: no cover
        if not isinstance(link_addr, str):
            raise ValueError("link_addr must be an NNG address string.") # pragma: no cover
        if not isinstance(host_addr, str):
            raise ValueError("host_addr must be an NNG address string.") # pragma: no cover
        if rank == 0 and link_addr != host_addr:
            raise ValueError(f"link_addr {link_addr} and host_addr {host_addr} must match for rank 0.") # pragma: no cover

        # Setup the player's receiving socket.
        self._receiver = pynng.Rep0(listen=host_addr, recv_timeout=nng_timeout(setup_timeout))
        log.info(f"Player {rank} rendezvous with {link_addr} from {host_addr}.")

        # Rank 0 waits for every player to send their address.
        if rank == 0:
            remaining_ranks = set(range(1, world_size))
            addresses = [(rank, host_addr)]

            for index in range(1, world_size):
                other_rank, other_host_addr, other_token = pickle.loads(self._receiver.recv())
                self._receiver.send(b"ok")

                if other_token != token:
                    raise RuntimeError(f"Player {rank} expected token {token}, received {other_token} from player {other_rank}.")
                addresses.append((other_rank, other_host_addr))

            # Setup sockets for sending to the other players.
            addresses = [address for rank, address in sorted(addresses)]
            self._players = [pynng.Req0(dial=address) for address in addresses]

            # Send addresses back to the other players.
            for player in self._players[1:]:
                player.send(pickle.dumps(addresses))
                player.recv()

        # All players send their address to rank 0.
        if rank != 0:
            with pynng.Req0(dial=link_addr, send_timeout=nng_timeout(setup_timeout)) as link:
                link.send(pickle.dumps((rank, host_addr, token)))
                link.recv()

            addresses = pickle.loads(self._receiver.recv())
            self._receiver.send(b"ok")

            # Setup sockets for sending to the other players.
            self._players = [pynng.Req0(dial=address) for address in addresses]

        # We don't want a timeout for the receiving socket.
        self._receiver.recv_timeout = nng_timeout(None)

        self._name = name
        self._world_size = world_size
        self._link_addr = link_addr
        self._rank = rank
        self._host_addr = host_addr
        self._timeout = timeout
        self._revoked = False

        self._send_serial = 0

        self._stats = {
             "bytes": {
                 "sent": {
                     "total": 0,
                 },
                 "received": {
                     "total": 0,
                 },
             },
            "messages": {
                "sent": {
                    "total": 0,
                },
                "received": {
                    "total": 0,
                    },
                },
        }

        # Setup queues for incoming messages.
        self._incoming = queue.Queue()
        self._receive_queues = {}
        for tag in self._tags:
            if tag not in ["revoke"]:
                self._receive_queues[tag] = {}
                for rank in self.ranks:
                    self._receive_queues[tag][rank] = queue.Queue()

        # Start queueing incoming messages.
        self._queueing_thread = threading.Thread(name="Queueing", target=self._queue_messages, daemon=True)
        self._queueing_thread.start()

        # Start receiving incoming messages.
        self._receiving_thread = threading.Thread(name="Incoming", target=self._receive_messages, daemon=True)
        self._receiving_thread.start()

        self._freed = False

        log.info(f"Comm {self.name!r} player {self._rank} communicator ready.")

    def _queue_messages(self):
        # Place incoming messages in the correct queue.
        while True:
            # Wait for the next incoming message.
            message = self._incoming.get(block=True, timeout=None)

            # If the communicator has been freed, exit the thread.
            if isinstance(message, NNGCommunicator._Done):
                return

            # Drop messages with missing attributes or unexpected values.
            try:
                if not hasattr(message, "payload"):
                    raise RuntimeError(f"Message missing payload.") # pragma: no cover
                if not hasattr(message, "sender"):
                    raise RuntimeError(f"Message missing sender.") # pragma: no cover
                if not hasattr(message, "serial"):
                    raise RuntimeError(f"Message missing serial number.") # pragma: no cover
                if not hasattr(message, "tag"):
                    raise RuntimeError(f"Message missing tag.") # pragma: no cover
                if message.tag not in self._tags:
                    raise RuntimeError(f"Unknown tag: {message.tag}") # pragma: no cover
                if message.sender not in self.ranks:
                    raise RuntimeError(f"Unexpected sender: {message.sender}") # pragma: no cover
            except Exception as e: # pragma: no cover
                log.error(f"Comm {self.name!r} player {self.rank} dropping unexpected message: {e}")
                continue

            # Log received messages.
            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"Comm {self.name!r} player {self.rank} <-- player {message.sender} {message.tag}#{message.serial:04}")

            # Revoke messages don't get queued because they receive special handling.
            if message.tag == "revoke":
                if not self._revoked:
                    self._revoked = True
                    log.info(f"Comm {self.name!r} player {self.rank} revoked by player {message.sender}")
                continue

            # Insert the message into the correct queue.
            self._receive_queues[message.tag][message.sender].put(message, block=True, timeout=None)


    def _receive_messages(self):
        # Parse and queue incoming messages as they arrive.
        while True:
            try:
                # Wait for a message to arrive from the pynng socket.
                raw_message = self._receiver.recv(block=True)
                self._receiver.send(b"ok")

                # Update statistics.
                self._stats["bytes"]["received"]["total"] += len(raw_message)
                self._stats["messages"]["received"]["total"] += 1

                # Ignore unparsable messages.
                try:
                    message = pickle.loads(raw_message)
                except Exception as e: # pragma: no cover
                    log.error(f"Comm {self.name!r} player {self.rank} ignoring unparsable message: {e}")
                    continue

                # Insert the message into the incoming queue.
                self._incoming.put(message, block=True, timeout=None)
            except pynng.exceptions.Closed:
                # The communicator has been freed, so exit the thread.
                log.debug(f"Comm {self.name!r} player {self.rank} receiving socket closed.")
                break


    def _receive(self, *, tag, sender, block):
        try:
            return self._receive_queues[tag][sender].get(block=block, timeout=self._timeout)
        except queue.Empty:
            if block:
                raise Timeout(f"Tag {tag!r} from sender {sender} timed-out after {self._timeout}s")
            else:
                raise TryAgain(f"Tag {tag!r} from sender {sender} try again.")


    def _require_rank(self, rank):
        if not isinstance(rank, numbers.Integral):
            raise ValueError("Rank must be an integer.") # pragma: no cover
        if rank < 0 or rank >= self._world_size:
            raise ValueError(f"Rank must be in the range [0, {self._world_size}).") # pragma: no cover
        return int(rank)


    def _require_rank_list(self, ranks):
        ranks = [self._require_rank(rank) for rank in ranks]
        if len(ranks) != len(set(ranks)):
            raise ValueError("Duplicate ranks are not allowed.") # pragma: no cover
        return ranks


    def _require_value(self, value):
        return value


    def _require_optional_value(self, value):
        return value


    def _require_unrevoked(self):
        if self._revoked:
            raise Revoked(f"Comm {self.name!r} player {self.rank} revoked.")


    def _send(self, *, tag, payload, dst):
        if tag not in self._tags:
            raise ValueError(f"Unknown tag: {tag}") # pragma: no cover
        if dst not in self.ranks:
            raise ValueError(f"Unknown destination: {dst}") # pragma: no cover

        message = Message(self._send_serial, tag, self._rank, payload)
        self._send_serial += 1

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"Comm {self.name!r} player {self.rank} --> player {dst} {message.tag}#{message.serial:04}") # pragma: no cover

        # As a special-case, route messages sent to ourself directly to the incoming queue.
        if dst == self.rank:
            self._incoming.put(message, block=True, timeout=None)
        # Otherwise, send the message to the appropriate socket.
        else:
            try:
                raw_message = pickle.dumps(message)
                player = self._players[dst]
                player.send_timeout = nng_timeout(self._timeout)
                player.send(raw_message)
                player.recv()
                self._stats["bytes"]["sent"]["total"] += len(raw_message)
                self._stats["messages"]["sent"]["total"] += 1
            except pynng.exceptions.Timeout:
                log.error(f"Comm {self.name!r} player {self.rank} tag {message.tag!r} to receiver {dst} timed-out after {self._timeout}s.")
            except Exception as e:
                log.error(f"Comm {self.name!r} player {self.rank} exception: {e}")


    def all_gather(self, value):
        log.debug(f"Comm {self.name!r} player {self.rank} all_gather()")

        self._require_unrevoked()
        value = self._require_value(value)

        # Send messages.
        for rank in self.ranks:
            self._send(tag="allgather", payload=value, dst=rank)

        # Receive messages.
        messages = [self._receive(tag="allgather", sender=rank, block=True) for rank in self.ranks]
        values = [message.payload for message in messages]
        return values


    def barrier(self):
        """If the implementation returns without raising an exception, then
        every player entered the barrier.  If an exception is raised then there
        are no guarantees about whether every player entered.
        """
        log.debug(f"Comm {self.name!r} player {self.rank} barrier()")

        self._require_unrevoked()

        # Notify rank 0 that we've entered the barrier.
        self._send(tag="barrier-enter", payload=None, dst=0)

        if self.rank == 0:
            # Wait until every player enters the barrier.
            for rank in self.ranks:
                self._receive(tag="barrier-enter", sender=rank, block=True)
            # Notify every player that it's time to exit the barrier.
            for rank in self.ranks:
                self._send(tag="barrier-exit", payload=None, dst=rank)

        # Wait until we're told to exit.
        self._receive(tag="barrier-exit", sender=0, block=True)


    def broadcast(self, *, src, value):
        log.debug(f"Comm {self.name!r} player {self.rank} broadcast(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)
        value = self._require_optional_value(value)

        # Broadcast the value to all players.
        if self.rank == src:
            for rank in self.ranks:
                self._send(tag="broadcast", payload=value, dst=rank)

        # Receive the broadcast value.
        message = self._receive(tag="broadcast", sender=src, block=True)
        return message.payload


    def free(self):
        # Calling free() multiple times is a no-op.
        if self._freed:
            return
        self._freed = True

        # Stop receiving.
        self._receiver.close()
        self._receiver = None
        # The following blocks intermittently when doing CI with Github
        # actions, for reasons I don't understand.
        #self._receiving_thread.join()

        # Stop handling incoming messages.
        self._incoming.put(NNGCommunicator._Done())
        self._queueing_thread.join()

        # Close outgoing sockets.
        for player in self._players:
            player.close()
        self._players = None

        log.info(f"Comm {self.name!r} player {self.rank} communicator freed.")


    def gather(self, *, value, dst):
        log.debug(f"Comm {self.name!r} player {self.rank} gather(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        # Send local data to the destination.
        self._send(tag="gather", payload=value, dst=dst)

        # Receive data from all ranks.
        if self.rank == dst:
            # Messages could arrive out of order.
            messages = [self._receive(tag="gather", sender=rank, block=True) for rank in self.ranks]
            values = [message.payload for message in messages]
            return values

        return None


    def gatherv(self, *, src, value, dst):
        log.debug(f"Comm {self.name!r} player {self.rank} gatherv(src={src}, dst={dst})")

        self._require_unrevoked()
        src = self._require_rank_list(src)
        value = self._require_value(value)
        dst = self._require_rank(dst)

        # Send local data to the destination.
        if self.rank in src:
            self._send(tag="gatherv", payload=value, dst=dst)

        # Receive data from the other players.
        if self.rank == dst:
            # Messages could arrive out of order.
            messages = [self._receive(tag="gatherv", sender=rank, block=True) for rank in src]
            values = [message.payload for message in messages]
            return values

        return None


    def irecv(self, *, src):
        log.debug(f"Comm {self.name!r} player {self.rank} irecv(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)

        class Result:
            def __init__(self, communicator, sender):
                self._communicator = communicator
                self._sender = sender
                self._message = None

            def is_completed(self):
                if self._message is None:
                    try:
                        self._message = self._communicator._receive(tag="send", sender=self._sender, block=False)
                    except TryAgain:
                        pass
                return self._message is not None

            @property
            def value(self):
                if self._message is None:
                    raise RuntimeError("Call not completed.")
                return self._message.payload

            def wait(self):
                if self._message is None:
                    self._message = self._communicator._receive(tag="send", sender=self._sender, block=True)

        return Result(self, src)


    def isend(self, *, value, dst):
        log.debug(f"Comm {self.name!r} player {self.rank} isend(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        self._send(tag="send", payload=value, dst=dst)

        # This is safe, because we pickle the value before returning; thus,
        # nothing the caller can do to the value will have unexpected
        # side-effects.
        class Result:
            def is_completed(self):
                return True
            def wait(self):
                pass

        return Result()


    def log_stats(self):
        """Log statistics about the number and size of messages handled by this communicator."""
        messages_sent = self._stats["messages"]["sent"]["total"]
        messages_received = self._stats["messages"]["received"]["total"]
        bytes_sent = self._stats["bytes"]["sent"]["total"]
        bytes_received = self._stats["bytes"]["received"]["total"]

        log.info(f"Comm {self.name!r} player {self._rank} sent {messages_sent} messages / {bytes_sent} bytes, received {messages_received} messages / {bytes_received} bytes.")


    @property
    def name(self):
        """The name of this communicator, which can be used for logging / debugging.

        Returns
        -------
        name: string
        """
        return self._name


    @property
    def rank(self):
        return self._rank


    def recv(self, *, src):
        log.debug(f"Comm {self.name!r} player {self.rank} recv(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)

        message = self._receive(tag="send", sender=src, block=True)
        return message.payload


    def revoke(self):
        """Revoke the current communicator.

        Revokes the communicator for this player, and any players able to
        receive messages.  A revoked communicator cannot be used to perform any
        operation other than :meth:`shrink`.  Typically, revoke should be
        called by any player that detects a communication failure, to initiate
        a recovery phase.
        """
        log.debug(f"Comm {self.name!r} player {self.rank} revoke()")

        # Notify all players that the communicator is revoked.
        for rank in self.ranks:
            try:
                self._send(tag="revoke", payload=None, dst=rank)
            except Exception as e:
                # We handle this here instead of propagating it to the
                # application layer because we expect some recipients to be
                # missing, else there'd be no reason to call revoke().
                log.error(f"Comm {self.name!r} player {self.rank} timeout revoking player {rank}.")


    @staticmethod
    def run(world_size):
        """Decorator for functions that should be run in parallel using sub-processes on the local host.

        This is extremely useful for running examples and regression tests on one machine.

        The decorated function *must* accept a communicator as its first
        argument.  Additional positional and keyword arguments may follow the
        communicator.

        To run computations across multiple hosts, you should use the
        :ref:`cicada-exec` command line executable instead.

        Parameters
        ----------
        world_size: integer, required
            The number of players to run the decorated function.

        Returns
        -------
        results: list
            The return value from the decorated function for each player, in
            rank order.  If a player process terminates unexpectedly, the
            result will be an instance of :class:`Terminated`, which can be
            used to access the process exit code.  If the player process raises
            a Python exception, the result will be an instance of
            :class:`Failed`, which can be used to access the Python exception
            and a traceback of the failing code.
        """
        def launch(*, world_size, link_addr, rank, host_addr, queue, func, args, kwargs):
            log = logging.getLogger(__name__)
            communicator = NNGCommunicator(world_size=world_size, link_addr=link_addr, rank=rank, host_addr=host_addr)
            try:
                result = func(communicator, *args, **kwargs)
                communicator.free()
            except Exception as e: # pragma: no cover
                result = Failed(e, traceback.format_exc())
            queue.put((rank, result))


        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                addresses = []
                for rank in range(world_size):
                    addr = cicada.bind.loopback_ip()
                    port = cicada.bind.random_port(addr)
                    addresses.append(f"tcp://{addr}:{port}")

                context = multiprocessing.get_context(method="fork")

                queue = context.Queue()

                # Start per-player processes.
                processes = []
                for rank in range(world_size):
                    processes.append(context.Process(target=launch, kwargs=dict(world_size=world_size, link_addr=addresses[0], rank=rank, host_addr=addresses[rank], queue=queue, func=func, args=args, kwargs=kwargs)))
                for process in processes:
                    process.daemon = True
                    process.start()

                # Wait until every process terminates.
                for process in processes:
                    process.join()

                # Collect a result for every process, but don't block in case
                # there are missing results.
                results = []
                for process in processes:
                    try:
                        results.append(queue.get(block=False))
                    except:
                        break

                # Return the output of each rank, in rank order, with a sentinel object for missing outputs.
                output = [Terminated(process.exitcode) for process in processes]
                for rank, result in results:
                    output[rank] = result

                # Log the results for each player.
                for rank, result in enumerate(output):
                    if isinstance(result, Failed):
                        log.error(f"Player {rank} failed: {result.exception!r}")
                    elif isinstance(result, Exception):
                        log.error(f"Player {rank} failed: {result!r}")
                    else:
                        log.info(f"Player {rank} returned: {result}")

                # Print a traceback for players that failed.
                for rank, result in enumerate(output):
                    if isinstance(result, Failed):
                        log.error("*" * 80)
                        log.error(f"Player {rank} traceback:")
                        log.error(result.traceback)

                return output


            return wrapper
        return decorator


    def scatter(self, *, src, values):
        log.debug(f"Comm {self.name!r} player {self.rank} scatter(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)
        if self.rank == src:
            values = [self._require_value(value) for value in values]
            if len(values) != self._world_size:
                raise ValueError(f"Expected {self._world_size} values, received {len(values)}.") # pragma: no cover

        # Send data to every player.
        if self.rank == src:
            for value, rank in zip(values, self.ranks):
                self._send(tag="scatter", payload=value, dst=rank)

        # Receive data from the sender.
        message = self._receive(tag="scatter", sender=src, block=True)
        return message.payload


    def scatterv(self, *, src, values, dst):
        log.debug(f"Comm {self.name!r} player {self.rank} scatterv(src={src}, dst={dst})")

        self._require_unrevoked()
        src = self._require_rank(src)
        dst = self._require_rank_list(dst)

        if self.rank == src:
            values = [self._require_value(value) for value in values]
            if len(values) != len(dst):
                raise ValueError("values must contain one value instance for every destination player.") # pragma: no cover

        # Send data to every destination.
        if self.rank == src:
            for value, rank in zip(values, dst):
                self._send(tag="scatterv", payload=value, dst=rank)

        # Receive data from the sender.
        if self.rank in dst:
            message = self._receive(tag="scatterv", sender=src, block=True)
            return message.payload

        return None


    def send(self, *, value, dst):
        log.debug(f"Comm {self.name!r} player {self.rank} send(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        self._send(tag="send", payload=value, dst=dst)


    def shrink(self, timeout=None, name=None):
        """Create a new communicator containing surviving players.

        This method should be called as part of a failure-recovery phase by as
        many players as possible (ideally, every player still running).  It
        will attempt to rendezvous with the other players and return a new
        communicator, but the process could fail and raise an exception
        instead.  In that case it is up to the application to decide how to
        proceed.
        """
        log.debug(f"Comm {self.name!r} player {self.rank} shrink()")

        # Set a default timeout of 2 seconds.
        if timeout is None:
            timeout = 2

        # Create a default name for the new communicator.
        if name is None:
            name = f"shrunk_{self.name}"

        # Our goal is to identify which players still exist.
        remaining_ranks = set()

        # Notify players that we're alive.
        for rank in self.ranks:
            self._send(tag="shrink-enter", payload=None, dst=rank)

        # Collect notifications from the other players during a window of time.
        start = time.time()
        while True:
            for rank in self.ranks:
                try:
                    message = self._receive(tag="shrink-enter", sender=rank, block=False)
                    remaining_ranks.add(rank)
                except Exception as e:
                    log.debug(f"Comm {self.name!r} player {self.rank} exception {e}")
            if time.time() - start > timeout:
                break
            time.sleep(0.1)

        # Sort the remaining ranks; the lowest rank will become rank 0 in the new communicator.
        remaining_ranks = sorted(list(remaining_ranks))
        #log.info(f"Comm {self.name!r} player {self.rank} remaining players: {remaining_ranks}")

        # Generate a token based on a hash of the remaining ranks that we can
        # use to ensure that every player is in agreement on who's remaining.
        token = hashlib.sha3_256()
        for rank in remaining_ranks:
            token.update(rank.to_bytes(math.ceil(rank.bit_length() / 8), byteorder="big"))
        token = token.hexdigest()

        # Generate new connection information.
        new_world_size=len(remaining_ranks)
        new_rank = remaining_ranks.index(self.rank)

        protocol, addr = self._host_addr.split("//")
        addr, port = addr.split(":")
        port = cicada.bind.random_port(addr)
        new_host_addr = f"{protocol}//{addr}:{port}"

        if self.rank == remaining_ranks[0]:
            for remaining_rank in remaining_ranks:
                self._send(tag="shrink-exit", payload=new_host_addr, dst=remaining_rank)
        new_link_addr = self._receive(tag="shrink-exit", sender=remaining_ranks[0], block=True).payload

        return NNGCommunicator(name=name, world_size=new_world_size, rank=new_rank, link_addr=new_link_addr, host_addr=new_host_addr, token=token), remaining_ranks


    def split(self, *, group):
        """Partition players into groups, and create new communicators for each.

        Each player supplies a group identifier - which can be any hashable
        Python type (string, integer, etc) - and a new communicator is created
        for each distinct group, with those players as members.  If a player
        supplies `None`, they will not be a part of any group, and this
        method will return `None`.

        .. note::
            This is a collective operation that *must* be called by every member
            of the communicator, even if they aren't going to be a member of any
            of the resulting groups!

        Parameters
        ----------
        group: hashable object, required
            Group identifier, or `None`.

        Returns
        -------
        communicator: a new :class:`Communicator` instance, or `None`
        """
        log.debug(f"Comm {self.name!r} player {self.rank} split(group={group})")

        self._require_unrevoked()

        # Generate a new address.
        protocol, addr = self._host_addr.split("//")
        addr, port = addr.split(":")
        port = cicada.bind.random_port(addr)
        host_addr = f"{protocol}//{addr}:{port}"

        # Send group membership and new address to rank 0.
        my_group = group
        my_host_addr = host_addr
        self._send(tag="split-prepare", payload=(my_group, my_host_addr), dst=0)

        if self.rank == 0:
            # Collect group membership information and new addresses from all players.
            messages = [self._receive(tag="split-prepare", sender=rank, block=True) for rank in self.ranks]
            groups = [message.payload[0] for message in messages]
            host_addrs = [message.payload[1] for message in messages]

            # Identify world_size and link_addr for each group.
            world_sizes = collections.Counter()
            link_addrs = {}
            new_ranks = []
            for group, host_addr in zip(groups, host_addrs):
                new_ranks.append(world_sizes[group])
                world_sizes[group] += 1
                if group not in link_addrs:
                    link_addrs[group] = host_addr

            # Send world_size and link_addr to all players.
            for rank, group, new_rank in zip(self.ranks, groups, new_ranks):
                self._send(tag="split", payload=(world_sizes[group], link_addrs[group], new_rank), dst=rank)

        # Collect group information.
        world_size, link_addr, new_rank = self._receive(tag="split", sender=0, block=True).payload

        # Return a new communicator.
        if my_group is not None:
            return NNGCommunicator(name=my_group, world_size=world_size, rank=new_rank, link_addr=link_addr, host_addr=my_host_addr)


    @property
    def stats(self):
        """Nested dict containing communication statistics for logging / debugging."""
        return self._stats

    @property
    def timeout(self):
        """Amount of time allowed to elapse before blocking communications time-out.

        Returns
        -------
        timeout: number or `None`.
            The timeout in seconds, or `None` if there is no timeout.
        """
        return self._timeout


    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout


    @contextlib.contextmanager
    def override(self, *, timeout):
        """Temporarily change send / receive timeout value.

        Parameters
        ----------
        timeout: number or `None`
            The timeout for subsequent send / receive operations, in seconds, or `None` to disable timeouts completely.

        Returns
        -------
        A context manager object that will restore the original timeout value when exited.
        """
        original_timeout, self._timeout = self._timeout, timeout
        try:
            yield original_timeout
        finally:
            self._timeout = original_timeout


    @property
    def world_size(self):
        return self._world_size





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for communicating using the builtin :mod:`socket` module.
"""

import collections
import contextlib
import functools
import hashlib
import logging
import math
import multiprocessing
import numbers
import os
import pickle
import queue
import select
import socket
import threading
import time
import traceback
import urllib.parse

import numpy
import pynetstring

from .interface import Communicator
import cicada
import cicada.bind

log = logging.getLogger(__name__)

Message = collections.namedtuple("Message", ["serial", "tag", "sender", "payload"])
Message.__doc__ = """
Wrapper class for messages sent between processes.
"""

class Failed(Exception):
    """Used to indicate that a player process raised an exception."""
    def __init__(self, exception, traceback):
        self.exception = exception
        self.traceback = traceback

    def __repr__(self):
        return f"Failed(exception={self.exception!r})" # pragma: no cover


class LoggerAdapter(logging.LoggerAdapter):
    """Wraps a Python logger for consistent formatting of communicator log entries.

    logger: :class:`logging.Logger`, required
        Python logger to wrap.
    name: class:`str`, required
        Communicator name.
    rank: class:`int`, required
        Communicator rank
    """
    def __init__(self, logger, name, rank):
        super().__init__(logger, extra={"name": name, "rank": rank})

    def process(self, msg, kwargs):
        return f"Comm {self.extra['name']!r} player {self.extra['rank']} {msg}", kwargs


class NetstringSocket(object):
    """Message-oriented socket that uses the Netstrings protocol."""
    def __init__(self, sock):
        self._socket = sock
        self._decoder = pynetstring.Decoder()
        self._decoded = []
        self._sent_bytes = 0
        self._sent_messages = 0
        self._received_bytes = 0
        self._received_messages = 0

    def close(self):
        """Close the underlying socket."""
        self._socket.close()

    def feed(self):
        """Read data from the underlying socket, decoding whatever is available."""
        raw = self._socket.recv(4096)
        decoded = self._decoder.feed(raw)
        self._received_bytes += len(raw)
        self._received_messages += len(decoded)
        self._decoded += decoded

    def fileno(self):
        """Return the file descriptor for the underlying socket.

        This allows :class:`NetstringSocket` to be used with :func:`select.select`.
        """
        return self._socket.fileno()

    def received(self):
        """Return every message that has been received, if any."""
        result = self._decoded
        self._decoded = []
        return result

    def receive_one(self):
        """Block until at least one message has been received."""
        while not self._decoded:
            self.feed()
        return self._decoded.pop(0)

    def send(self, msg):
        """Send a message."""
        raw = pynetstring.encode(msg)
        self._socket.sendall(raw)
        self._sent_bytes += len(raw)
        self._sent_messages += 1

    @property
    def stats(self):
        """Return a :class:`dict` containing statistics.

        Returns the number of bytes and messages that have been sent and received.
        """
        return {
            "sent": {"bytes": self._sent_bytes, "messages": self._sent_messages},
            "received": {"bytes": self._received_bytes, "messages": self._received_messages},
            }


class Revoked(Exception):
    """Raised calling an operation after the communicator has been revoked."""
    pass


class Terminated(Exception):
    """Used to indicate that a player process terminated unexpectedly without output."""
    def __init__(self, exitcode):
        self.exitcode = exitcode

    def __repr__(self):
        return f"Terminated(exitcode={self.exitcode!r})" # pragma: no cover


class Timeout(Exception):
    """Raised when an operation has timed-out."""
    pass


class Timer(object):
    """Tracks elapsed time.

    Parameters
    ----------
    threshold: number, required
        The maximum of number of elapsed seconds before this timer will be expired.
    """
    def __init__(self, threshold):
        self._start = time.time()
        self._threshold = threshold

    @property
    def elapsed(self):
        """Return the number of seconds since the :class:`Timer` was created."""
        return time.time() - self._start


    @property
    def expired(self):
        """Returns :any:`True` if the elapsed time has exceeded a threshold."""
        if self._threshold and (self.elapsed > self._threshold):
            return True
        return False


class TryAgain(Exception):
    """Raised when a non-blocking operation would block."""
    pass


class SocketCommunicator(Communicator):
    """Cicada communicator that uses Python's builtin :mod:`socket` module as the transport layer.

    Note
    ----
    Creating a communicator is a collective operation that must be called by
    all players that will be members.

    Parameters
    ----------
    name: string, optional
        The name of this communicator, which is used strictly for logging
        and debugging.  If unspecified the default is "world".
    world_size: integer, optional
        The number of players that will be members of this communicator.
        Defaults to the value of the WORLD_SIZE environment variable.
    link_addr: string, optional
        URL address of the root (rank 0) player.  The URL scheme *must* be
        `tcp`, and the address must be reachable by all of the other players.
        Defaults to the value of the LINK_ADDR environment variable.
    rank: integer, optional
        The rank of the local player, in the range [0, world_size).  Defaults
        to the value of the RANK environment variable.
    host_addr: string, optional
        URL address of the local player.  The URL scheme *must* be `tcp` and
        the address must be reachable by all of the other players.  Defaults to
        the address of host_socket if supplied, or the value of the HOST_ADDR
        environment variable.  Note that this value is ignored by the root
        player.
    host_socket: :class:`socket.socket`, optional
        Callers may optionally provide an existing socket for use by the
        communicator.  The provided socket *must* be created using `AF_INET`
        and `SOCK_STREAM`, and be bound to the same address and port specified
        by `host_addr`.
    timeout: number or `None`
        Maximum time to wait for normal communication to complete in seconds, or `None` to disable timeouts.
    setup_timeout: number or `None`
        Maximum time allowed to setup the communicator in seconds, or `None` to disable timeouts during setup.
    """

    _tags = [
        "allgather",
        "barrier-enter",
        "barrier-exit",
        "broadcast",
        "gather",
        "gatherv",
        "revoke",
        "scatter",
        "scatterv",
        "send",
        "shrink-enter",
        "shrink-exit",
        "split",
        "split-prepare",
        ]


    def __init__(self, *, name=None, world_size=None, rank=None, link_addr=None, host_addr=None, host_socket=None, token=0, timeout=5, setup_timeout=5):
        # Enforce preconditions
        if name is None:
            name = "world"
        if not isinstance(name, str):
            raise ValueError("name must be a string, got {name} instead.") # pragma: no cover

        if world_size is None:
            world_size = int(os.environ["WORLD_SIZE"])
        if not isinstance(world_size, int):
            raise ValueError("world_size must be an integer, got {world_size} instead.") # pragma: no cover
        if not world_size > 0:
            raise ValueError("world_size must be an integer greater than zero, got {world_size} instead.") # pragma: no cover

        if rank is None:
            rank = int(os.environ["RANK"])
        if not isinstance(rank, int):
            raise ValueError("rank must be an integer, got {rank} instead.") # pragma: no cover
        if not (0 <= rank and rank < world_size):
            raise ValueError(f"rank must be in the range [0, {world_size}), got {rank} instead.") # pragma: no cover

        if link_addr is None:
            link_addr = os.environ["LINK_ADDR"]
        link_addr = urllib.parse.urlparse(link_addr)
        if link_addr.scheme != "tcp":
            raise ValueError("link_addr scheme must be tcp, got {link_addr.scheme} instead.") # pragma: no cover
        if link_addr.hostname is None:
            raise ValueError("link_addr hostname must be specified.") # pragma: no cover
        if link_addr.port is None:
            raise ValueError("link_addr port must be specified.") # pragma: no cover

        if host_addr is not None and host_socket is not None:
            raise ValueError("Specify host_addr or host_socket, but not both.") # pragma: no cover
        if host_addr is None and host_socket is not None:
            hostname, port = host_socket.getsockname()
            host_addr = f"tcp://{hostname}:{port}"
        if host_addr is None and host_socket is None:
            host_addr = os.environ["HOST_ADDR"]
        host_addr = urllib.parse.urlparse(host_addr)
        if host_addr.scheme != "tcp":
            raise ValueError("host_addr scheme must be tcp, got {host_addr.scheme} instead.") # pragma: no cover
        if host_addr.hostname is None:
            raise ValueError("host_addr hostname must be specified.") # pragma: no cover
        if host_addr.port is None:
            raise ValueError("host_addr port must be specified.") # pragma: no cover
        if rank == 0 and host_addr != link_addr:
            raise ValueError(f"Player 0 link_addr {link_addr} and host_addr {host_addr} must match.") # pragma: no cover

        if not isinstance(host_socket, (socket.socket, type(None))):
            raise ValueError(f"host_socket must be an instance of socket.socket or None.") # pragma: no cover
        if host_socket is not None and host_socket.family != socket.AF_INET:
            raise ValueError(f"host_socket must use AF_INET.") # pragma: no cover
        if host_socket is not None and host_socket.type != socket.SOCK_STREAM:
            raise ValueError(f"host_socket must use SOCK_STREAM.") # pragma: no cover
        if host_socket is not None and host_socket.getsockname()[0] != host_addr.hostname:
            raise ValueError(f"host_socket hostname must match host_addr.") # pragma: no cover
        if host_socket is not None and host_socket.getsockname()[1] != host_addr.port:
            raise ValueError(f"host_socket port must match host_addr.") # pragma: no cover

        if not isinstance(timeout, (numbers.Number, type(None))):
            raise ValueError(f"timeout must be a number or None, got {timeout} instead.") # pragma: no cover
        if not isinstance(setup_timeout, (numbers.Number, type(None))):
            raise ValueError(f"setup_timeout must be a number or None, got {setup_timeout} instead.") # pragma: no cover

        # Setup internal state.
        self._name = name
        self._world_size = world_size
        self._rank = rank
        self._host_addr = host_addr
        self._timeout = timeout
        self._setup_timeout = setup_timeout
        self._revoked = False
        self._log = LoggerAdapter(log, name, rank)

        # Set aside storage for connections to the other players.
        self._players = {}

        # Track elapsed time during setup.
        timer = Timer(threshold=setup_timeout)

        # Rendezvous with the other players.
        self._log.info(f"rendezvous with {link_addr.geturl()} from {host_addr.geturl()}.")

        ###########################################################################
        # Phase 1: Every player sets-up a socket to listen for connections.

        if host_socket is None:
            while True:
                if timer.expired:
                    raise Timeout(f"Comm {name!r} player {rank} timeout creating host socket.")

                try:
                    host_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    host_socket.bind((host_addr.hostname, host_addr.port or 0))
                    break
                except Exception as e:
                    self._log.warning(f"exception creating host socket: {e}")
                    time.sleep(0.1)

        host_socket.listen(world_size)
        self._log.debug(f"listening for connections.")

        ###########################################################################
        # Phase 2: Every player (except root) makes a connection to root.

        if rank != 0:
            while True:
                if timer.expired:
                    raise Timeout(f"Comm {name!r} player {rank} timeout connecting to player 0.")

                try:
                    other_player = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    other_player.connect((link_addr.hostname, link_addr.port))
                    other_player = NetstringSocket(other_player)
                    self._players[0] = other_player
                    break
                except Exception as e:
                    self._log.warning(f"exception connecting to player 0: {e}")
                    time.sleep(0.1)

        ###########################################################################
        # Phase 3: Every player sends their listening address to root.

        if rank != 0:
            while True:
                if timer.expired:
                    raise Timeout(f"Comm {name!r} player {rank} timeout sending address to player 0.")

                try:
                    self._players[0].send(pickle.dumps((rank, host_addr, token)))
                    break
                except Exception as e:
                    self._log.warning(f"exception sending address to player 0: {e}")
                    time.sleep(0.1)

        ###########################################################################
        # Phase 4: Root gathers addresses from every player.

        if rank == 0:
            # Gather an address from every player.
            addresses = {rank: host_addr}
            while len(addresses) < world_size:
                other_player, _ = host_socket.accept()
                other_player = NetstringSocket(other_player)
                other_rank, other_addr, other_token = pickle.loads(other_player.receive_one())
                self._players[other_rank] = other_player
                addresses[other_rank] = other_addr
                self._log.debug(f"received address from player {other_rank}.")

#                if other_token != token:
#                    raise RuntimeError(f"Comm {self._name!r} player {self._rank} expected token {token}, received {other_token} from player {other_rank}.")

        ###########################################################################
        # Phase 5: Root broadcasts the set of all addresses to every player.

        if rank == 0:
            for player in self._players.values():
                player.send(pickle.dumps(addresses))

        ###########################################################################
        # Phase 6: Every player receives the set of all addresses from root.

        if rank != 0:
            while True:
                if timer.expired:
                    raise Timeout(f"Comm {name!r} player {rank} timeout waiting for addresses from player 0.")

                try:
                    addresses = pickle.loads(self._players[0].receive_one())
                    self._log.debug(f"received addresses from player 0.")
                    break
                except Exception as e:
                    self._log.warning(f"exception getting addresses from player 0: {e}")
                    time.sleep(0.1)

        ###########################################################################
        # Phase 7: Players setup connections with one another.

        for listener in range(1, world_size-1):
            if rank == listener:
                # Listen for connections from the other players.
                while len(self._players) < world_size - 1: # we don't make a connection with ourself.
                    try:
                        other_player, _ = host_socket.accept()
                        other_player = NetstringSocket(other_player)
                        other_rank = pickle.loads(other_player.receive_one())
                        self._players[other_rank] = other_player
                        self._players[other_rank].send("ack")
                    except Exception as e:
                        self._log.warning(f"exception listening for other players: {e}")
                        time.sleep(0.1)

            elif rank > listener:
                while True:
                    if timer.expired:
                        raise Timeout(f"Comm {name!r} player {rank} timeout connecting to player {listener}.")

                    try:
                        # Send our rank to the listening player and listen for an ack.
                        other_player = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        other_player.connect((addresses[listener].hostname, addresses[listener].port))
                        other_player = NetstringSocket(other_player)
                        self._players[listener] = other_player

                        self._players[listener].send(pickle.dumps(rank))
                        ack = self._players[listener].receive_one()
                        break
                    except Exception as e:
                        self._log.warning(f"exception connecting to player {listener}: {e}")
                        time.sleep(0.5)

        ###########################################################################
        # Phase 8: The mesh has been initialized, setup normal operation.

        self._send_serial = 0
        self._running = True

        # Setup queues for incoming messages.
        self._incoming = queue.Queue()
        self._receive_queues = {}
        for tag in self._tags:
            if tag not in ["revoke"]:
                self._receive_queues[tag] = {}
                for rank in self.ranks:
                    self._receive_queues[tag][rank] = queue.Queue()

        # Start queueing incoming messages.
        self._queueing_thread = threading.Thread(name="Queueing", target=self._queue_messages, daemon=True)
        self._queueing_thread.start()

        # Start receiving incoming messages.
        self._incoming_thread = threading.Thread(name="Incoming", target=self._receive_messages, daemon=True)
        self._incoming_thread.start()

#        # Log information about our peers.
#        for rank, player in sorted(self._players.items()):
#            host, port = player.getsockname()
#            otherhost, otherport = player.getpeername()
#            self._log.debug(f"tcp://{host}:{port} connected to player {rank} tcp://{otherhost}:{otherport}.")

        self._log.info(f"communicator ready.")

    def _queue_messages(self):
        # Place incoming messages in the correct queue.
        while self._running:
            # Wait for the next incoming message.
            try:
                message = self._incoming.get(block=True, timeout=0.1)
            except queue.Empty:
                continue

            # Drop messages with missing attributes or unexpected values.
            try:
                if not hasattr(message, "payload"):
                    raise RuntimeError(f"Message missing payload.") # pragma: no cover
                if not hasattr(message, "sender"):
                    raise RuntimeError(f"Message missing sender.") # pragma: no cover
                if not hasattr(message, "serial"):
                    raise RuntimeError(f"Message missing serial number.") # pragma: no cover
                if not hasattr(message, "tag"):
                    raise RuntimeError(f"Message missing tag.") # pragma: no cover
                if message.tag not in self._tags:
                    raise RuntimeError(f"Unknown tag: {message.tag}") # pragma: no cover
                if message.sender not in self.ranks:
                    raise RuntimeError(f"Unexpected sender: {message.sender}") # pragma: no cover
            except Exception as e: # pragma: no cover
                self._log.error(f"dropping unexpected message: {message} exception: {e}")
                continue

            # Log queued messages.
            if self._log.isEnabledFor(logging.DEBUG):
                self._log.debug(f"<-- player {message.sender} {message.tag}#{message.serial:04}")

            # Revoke messages don't get queued because they receive special handling.
            if message.tag == "revoke":
                if not self._revoked:
                    self._revoked = True
                    self._log.info(f"revoked by player {message.sender}")
                continue

            # Insert the message into the correct queue.
            self._receive_queues[message.tag][message.sender].put(message, block=True, timeout=None)

        self._log.debug(f"queueing thread closed.")


    def _receive_messages(self):
        # Parse and queue incoming messages as they arrive.
        while self._running:
            try:
                # Wait for data to arrive from the other players.
                ready, _, _ = select.select(self._players.values(), [], [], 0.01)
                for player in ready:
                    player.feed()

                # Process any messages that were received. Note that
                # we iterate over every player, not just the ones that
                # were selected above, because there might be a few
                # messages left from the startup process.
                for player in self._players.values():
                    for raw_message in player.received():
                        # Ignore unparsable messages.
                        try:
                            message = pickle.loads(raw_message)
                        except Exception as e: # pragma: no cover
                            self._log.error(f"ignoring unparsable message: {e}")
                            continue

                        self._log.debug(f"received {message}")

                        # Insert the message into the incoming queue.
                        self._incoming.put(message, block=True, timeout=None)
            except Exception as e:
                self._log.error(f"receive exception: {e}")

        # The communicator has been freed, so exit the thread.
        self._log.debug(f"receive thread closed.")



    def _receive(self, *, tag, sender, block):
        try:
            return self._receive_queues[tag][sender].get(block=block, timeout=self._timeout)
        except queue.Empty:
            if block:
                raise Timeout(f"Tag {tag!r} from sender {sender} timed-out after {self._timeout}s")
            else:
                raise TryAgain(f"Tag {tag!r} from sender {sender} try again.")


    def _require_rank(self, rank):
        if not isinstance(rank, numbers.Integral):
            raise ValueError("Rank must be an integer.") # pragma: no cover
        if rank < 0 or rank >= self._world_size:
            raise ValueError(f"Rank must be in the range [0, {self._world_size}).") # pragma: no cover
        return int(rank)


    def _require_rank_list(self, ranks):
        ranks = [self._require_rank(rank) for rank in ranks]
        if len(ranks) != len(set(ranks)):
            raise ValueError("Duplicate ranks are not allowed.") # pragma: no cover
        return ranks


    def _require_value(self, value):
        return value


    def _require_optional_value(self, value):
        return value


    def _require_unrevoked(self):
        if self._revoked:
            raise Revoked(f"Comm {self.name!r} player {self.rank} revoked.")


    def _send(self, *, tag, payload, dst):
        if tag not in self._tags:
            raise ValueError(f"Unknown tag: {tag}") # pragma: no cover
        if dst not in self.ranks:
            raise ValueError(f"Unknown destination: {dst}") # pragma: no cover

        message = Message(self._send_serial, tag, self._rank, payload)
        self._send_serial += 1

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug(f"--> player {dst} {message.tag}#{message.serial:04}") # pragma: no cover

        # As a special-case, route messages sent to ourself directly to the incoming queue.
        if dst == self.rank:
            self._incoming.put(message, block=True, timeout=None)
        # Otherwise, send the message to the appropriate socket.
        else:
            try:
                raw_message = pickle.dumps(message)
                player = self._players[dst]
                player.send(raw_message)
#            except pynng.exceptions.Timeout:
#                self._log.error(f"tag {message.tag!r} to receiver {dst} timed-out after {self._timeout}s.")
            except Exception as e:
                self._log.error(f"send exception: {e}")


    def all_gather(self, value):
        self._log.debug(f"all_gather()")

        self._require_unrevoked()
        value = self._require_value(value)

        # Send messages.
        for rank in self.ranks:
            self._send(tag="allgather", payload=value, dst=rank)

        # Receive messages.
        messages = [self._receive(tag="allgather", sender=rank, block=True) for rank in self.ranks]
        values = [message.payload for message in messages]
        return values


    def barrier(self):
        """If the implementation returns without raising an exception, then
        every player entered the barrier.  If an exception is raised then there
        are no guarantees about whether every player entered.
        """
        self._log.debug(f"barrier()")

        self._require_unrevoked()

        # Notify rank 0 that we've entered the barrier.
        self._send(tag="barrier-enter", payload=None, dst=0)

        if self.rank == 0:
            # Wait until every player enters the barrier.
            for rank in self.ranks:
                self._receive(tag="barrier-enter", sender=rank, block=True)
            # Notify every player that it's time to exit the barrier.
            for rank in self.ranks:
                self._send(tag="barrier-exit", payload=None, dst=rank)

        # Wait until we're told to exit.
        self._receive(tag="barrier-exit", sender=0, block=True)


    def broadcast(self, *, src, value):
        self._log.debug(f"broadcast(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)
        value = self._require_optional_value(value)

        # Broadcast the value to all players.
        if self.rank == src:
            for rank in self.ranks:
                self._send(tag="broadcast", payload=value, dst=rank)

        # Receive the broadcast value.
        message = self._receive(tag="broadcast", sender=src, block=True)
        return message.payload


    def free(self):
        # Calling free() multiple times is a no-op.
        if not self._running:
            return

        self._running = False

        # Stop receiving incoming messages.
        self._incoming_thread.join()

        # Stop queueing incoming messages.
        self._queueing_thread.join()

        # Close connections to the other players.
        for player in self._players.values():
            player.close()

        self._log.info(f"communicator freed.")


    def gather(self, *, value, dst):
        self._log.debug(f"gather(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        # Send local data to the destination.
        self._send(tag="gather", payload=value, dst=dst)

        # Receive data from all ranks.
        if self.rank == dst:
            # Messages could arrive out of order.
            messages = [self._receive(tag="gather", sender=rank, block=True) for rank in self.ranks]
            values = [message.payload for message in messages]
            return values

        return None


    def gatherv(self, *, src, value, dst):
        self._log.debug(f"gatherv(src={src}, dst={dst})")

        self._require_unrevoked()
        src = self._require_rank_list(src)
        value = self._require_value(value)
        dst = self._require_rank(dst)

        # Send local data to the destination.
        if self.rank in src:
            self._send(tag="gatherv", payload=value, dst=dst)

        # Receive data from the other players.
        if self.rank == dst:
            # Messages could arrive out of order.
            messages = [self._receive(tag="gatherv", sender=rank, block=True) for rank in src]
            values = [message.payload for message in messages]
            return values

        return None


    def irecv(self, *, src):
        self._log.debug(f"irecv(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)

        class Result:
            def __init__(self, communicator, sender):
                self._communicator = communicator
                self._sender = sender
                self._message = None

            def is_completed(self):
                if self._message is None:
                    try:
                        self._message = self._communicator._receive(tag="send", sender=self._sender, block=False)
                    except TryAgain:
                        pass
                return self._message is not None

            @property
            def value(self):
                if self._message is None:
                    raise RuntimeError("Call not completed.")
                return self._message.payload

            def wait(self):
                if self._message is None:
                    self._message = self._communicator._receive(tag="send", sender=self._sender, block=True)

        return Result(self, src)


    def isend(self, *, value, dst):
        self._log.debug(f"isend(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        self._send(tag="send", payload=value, dst=dst)

        # This is safe, because we pickle the value before returning; thus,
        # nothing the caller can do to the value will have unexpected
        # side-effects.
        class Result:
            def is_completed(self):
                return True
            def wait(self):
                pass

        return Result()


    @property
    def name(self):
        """The name of this communicator, which can be used for logging / debugging.

        Returns
        -------
        name: string
        """
        return self._name


    @contextlib.contextmanager
    def override(self, *, timeout):
        """Temporarily change the timeout value.

        Parameters
        ----------
        timeout: number or `None`
            The timeout for subsequent send / receive operations, in seconds, or `None` to disable timeouts completely.

        Returns
        -------
        A context manager object that will restore the original timeout value when exited.
        """
        original_timeout, self._timeout = self._timeout, timeout
        try:
            yield original_timeout
        finally:
            self._timeout = original_timeout


    @property
    def rank(self):
        return self._rank


    def recv(self, *, src):
        self._log.debug(f"recv(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)

        message = self._receive(tag="send", sender=src, block=True)
        return message.payload


    def revoke(self):
        """Revoke the current communicator.

        Revokes the communicator for this player, and any players able to
        receive messages.  A revoked communicator cannot be used to perform any
        operation other than :meth:`shrink`.  Typically, revoke should be
        called by any player that detects a communication failure, to initiate
        a recovery phase.
        """
        self._log.debug(f"revoke()")

        # Notify all players that the communicator is revoked.
        for rank in self.ranks:
            try:
                self._send(tag="revoke", payload=None, dst=rank)
            except Exception as e:
                # We handle this here instead of propagating it to the
                # application layer because we expect some recipients to be
                # missing, else there'd be no reason to call revoke().
                self._log.error(f"timeout revoking player {rank}.")


    @staticmethod
    def run(*, world_size, fn, args=(), kwargs={}, timeout=5, setup_timeout=5):
        """Run a function in parallel using sub-processes on the local host.

        This is extremely useful for running examples and regression tests on one machine.

        The given function *must* accept a communicator as its first
        argument.  Additional positional and keyword arguments may follow the
        communicator.

        To run computations across multiple hosts, you should use the
        :ref:`cicada-exec` command line executable instead.

        Parameters
        ----------
        world_size: :class:`int`, required
            The number of players that will run the function.
        fn: callable object, required
            The function to execute in parallel.
        args: :class:`tuple`, optional
            Positional arguments to pass to `fn` when it is executed.
        kwargs: :class:`dict`, optional
            Keyword arguments to pass to `fn` when it is executed.
        timeout: number or `None`
            Maximum time to wait for normal communication to complete in seconds, or `None` to disable timeouts.
        setup_timeout: number or `None`
            Maximum time allowed to setup the communicator in seconds, or `None` to disable timeouts during setup.

        Returns
        -------
        results: list
            The return value from the function for each player, in
            rank order.  If a player process terminates unexpectedly, the
            result will be an instance of :class:`Terminated`, which can be
            used to access the process exit code.  If the player process raises
            a Python exception, the result will be an instance of
            :class:`Failed`, which can be used to access the Python exception
            and a traceback of the failing code.
        """
        def launch(*, link_addr_queue, result_queue, world_size, rank, fn, args, kwargs, timeout, setup_timeout):
            # Create a socket with a randomly-assigned port number.
            host_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            host_socket.bind(("127.0.0.1", 0))
            host, port = host_socket.getsockname()
            host_addr = f"tcp://{host}:{port}"

            # Send address information to our peers.
            if rank == 0:
                for index in range(world_size):
                    link_addr_queue.put(host_addr)

            link_addr = link_addr_queue.get()

            # Run the work function.
            try:
                communicator = SocketCommunicator(world_size=world_size, link_addr=link_addr, rank=rank, host_socket=host_socket, timeout=timeout, setup_timeout=setup_timeout)
                result = fn(communicator, *args, **kwargs)
                communicator.free()
            except Exception as e: # pragma: no cover
                result = Failed(e, traceback.format_exc())

            # Return results to the parent process.
            result_queue.put((rank, result))

        # Setup the multiprocessing context.
        context = multiprocessing.get_context(method="fork") # I don't remember why we preferred fork().

        # Create queues for IPC.
        link_addr_queue = context.Queue()
        result_queue = context.Queue()

        # Create per-player processes.
        processes = []
        for rank in range(world_size):
            processes.append(context.Process(
                target=launch,
                kwargs=dict(link_addr_queue=link_addr_queue, result_queue=result_queue, world_size=world_size, rank=rank, fn=fn, args=args, kwargs=kwargs, timeout=timeout, setup_timeout=setup_timeout),
                ))

        # Start per-player processes.
        for process in processes:
            process.daemon = True
            process.start()

        # Wait until every process terminates.
        for process in processes:
            process.join()

        # Collect a result for every process, but don't block in case
        # there are missing results.
        results = []
        for process in processes:
            try:
                results.append(result_queue.get(block=False))
            except:
                break

        # Return the output of each rank, in rank order, with a sentinel object for missing outputs.
        output = [Terminated(process.exitcode) for process in processes]
        for rank, result in results:
            output[rank] = result

        # Log the results for each player.
        for rank, result in enumerate(output):
            if isinstance(result, Failed):
                log.info(f"Player {rank} failed: {result.exception!r}")
            elif isinstance(result, Exception):
                log.info(f"Player {rank} failed: {result!r}")
            else:
                log.info(f"Player {rank} return: {result}")

        # Print a traceback for players that failed.
        for rank, result in enumerate(output):
            if isinstance(result, Failed):
                log.error("*" * 80)
                log.error(f"Player {rank} traceback:")
                log.error(result.traceback)

        return output


    def scatter(self, *, src, values):
        self._log.debug(f"scatter(src={src})")

        self._require_unrevoked()
        src = self._require_rank(src)
        if self.rank == src:
            values = [self._require_value(value) for value in values]
            if len(values) != self._world_size:
                raise ValueError(f"Expected {self._world_size} values, received {len(values)}.") # pragma: no cover

        # Send data to every player.
        if self.rank == src:
            for value, rank in zip(values, self.ranks):
                self._send(tag="scatter", payload=value, dst=rank)

        # Receive data from the sender.
        message = self._receive(tag="scatter", sender=src, block=True)
        return message.payload


    def scatterv(self, *, src, values, dst):
        self._log.debug(f"scatterv(src={src}, dst={dst})")

        self._require_unrevoked()
        src = self._require_rank(src)
        dst = self._require_rank_list(dst)

        if self.rank == src:
            values = [self._require_value(value) for value in values]
            if len(values) != len(dst):
                raise ValueError("values must contain one value instance for every destination player.") # pragma: no cover

        # Send data to every destination.
        if self.rank == src:
            for value, rank in zip(values, dst):
                self._send(tag="scatterv", payload=value, dst=rank)

        # Receive data from the sender.
        if self.rank in dst:
            message = self._receive(tag="scatterv", sender=src, block=True)
            return message.payload

        return None


    def send(self, *, value, dst):
        self._log.debug(f"send(dst={dst})")

        self._require_unrevoked()
        value = self._require_value(value)
        dst = self._require_rank(dst)

        self._send(tag="send", payload=value, dst=dst)


    def shrink(self, timeout=None, name=None):
        """Create a new communicator containing surviving players.

        This method should be called as part of a failure-recovery phase by as
        many players as possible (ideally, every player still running).  It
        will attempt to rendezvous with the other players and return a new
        communicator, but the process could fail and raise an exception
        instead.  In that case it is up to the application to decide how to
        proceed.
        """
        self._log.debug(f"shrink()")

        # Set a default timeout of 2 seconds.
        if timeout is None:
            timeout = 2

        # Create a default name for the new communicator.
        if name is None:
            name = f"shrunk_{self.name}"

        # Our goal is to identify which players still exist.
        remaining_ranks = set()

        # Notify players that we're alive.
        for rank in self.ranks:
            self._send(tag="shrink-enter", payload=None, dst=rank)

        # Collect notifications from the other players during a window of time.
        start = time.time()
        while True:
            for rank in self.ranks:
                try:
                    message = self._receive(tag="shrink-enter", sender=rank, block=False)
                    remaining_ranks.add(rank)
                except Exception as e:
                    self._log.debug(f"exception {e}")
            if time.time() - start > timeout:
                break
            time.sleep(0.1)

        # Sort the remaining ranks; the lowest rank will become rank 0 in the new communicator.
        remaining_ranks = sorted(list(remaining_ranks))
        #self._log.info(f"remaining players: {remaining_ranks}")

        # Generate a token based on a hash of the remaining ranks that we can
        # use to ensure that every player is in agreement on who's remaining.
        token = hashlib.sha3_256()
        for rank in remaining_ranks:
            token.update(rank.to_bytes(math.ceil(rank.bit_length() / 8), byteorder="big"))
        token = token.hexdigest()

        # Generate new connection information.
        new_world_size=len(remaining_ranks)
        new_rank = remaining_ranks.index(self.rank)

        protocol, addr = self._host_addr.split("//")
        addr, port = addr.split(":")
        port = cicada.bind.random_port(addr)
        new_host_addr = f"{protocol}//{addr}:{port}"

        if self.rank == remaining_ranks[0]:
            for remaining_rank in remaining_ranks:
                self._send(tag="shrink-exit", payload=new_host_addr, dst=remaining_rank)
        new_link_addr = self._receive(tag="shrink-exit", sender=remaining_ranks[0], block=True).payload

        return SocketCommunicator(name=name, world_size=new_world_size, rank=new_rank, link_addr=new_link_addr, host_addr=new_host_addr, token=token), remaining_ranks


    def split(self, *, name):
        """Return a new communicator with the given name.

        If players specify different names - which can be any :class:`str`
        - then a new communicator will be created for each unique name, with
        those players as members.  If a player supplies a name of `None`,
        they will not be a part of any communicator, and this method will
        return `None`.

        .. note::
            This is a collective operation that *must* be called by every member
            of the communicator, even if they aren't going to be a member of any
            of the resulting groups!

        Parameters
        ----------
        name: :class:`str` or :any:`None`, required
            Communicator name, or `None`.

        Returns
        -------
        communicator: a new :class:`SocketCommunicator` instance, or `None`
        """
        self._log.debug(f"split(name={name})")

        self._require_unrevoked()

        if not isinstance(name, (str, type(None))):
            raise ValueError(f"name must be a string or None, got {name} instead.") # pragma: no cover

        # Create a new socket with a randomly-assigned port number.
        if name is not None:
            host_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            host_socket.bind((self._host_addr.hostname, 0))
            host, port = host_socket.getsockname()

            my_name = name
            my_host_addr = f"tcp://{host}:{port}"
        else:
            my_name = None
            my_host_addr = None

        # Send names and addresses to rank 0.
        self._send(tag="split-prepare", payload=(my_name, my_host_addr), dst=0)

        # Collect name membership information from all players and compute new communicator parameters.
        if self.rank == 0:
            messages = [self._receive(tag="split-prepare", sender=rank, block=True) for rank in self.ranks]
            new_names = [message.payload[0] for message in messages]
            new_host_addrs = [message.payload[1] for message in messages]

            new_world_sizes = collections.Counter()
            new_ranks = []
            for new_name in new_names:
                new_ranks.append(new_world_sizes[new_name])
                new_world_sizes[new_name] += 1

            new_world_sizes = [new_world_sizes[new_name] for new_name in new_names]

            new_link_addrs = {}
            for new_name, new_host_addr in zip(new_names, new_host_addrs):
                if new_name not in new_link_addrs:
                    new_link_addrs[new_name] = new_host_addr
            new_link_addrs = [new_link_addrs[new_name] for new_name in new_names]

        # Send name, world_size, rank, and link_addr to all players.
        if self.rank == 0:
            for dst, (new_name, new_world_size, new_rank, new_link_addr) in enumerate(zip(new_names, new_world_sizes, new_ranks, new_link_addrs)):
                self._send(tag="split", payload=(new_name, new_world_size, new_rank, new_link_addr), dst=dst)

        # Collect name information.
        new_name, new_world_size, new_rank, new_link_addr = self._receive(tag="split", sender=0, block=True).payload
        # Return a new communicator.
        if new_name is not None:
            return SocketCommunicator(name=new_name, world_size=new_world_size, rank=new_rank, link_addr=new_link_addr, host_socket=host_socket, timeout=self._timeout, setup_timeout=self._setup_timeout)

        return None


    @property
    def stats(self):
        """Nested dict containing communication statistics for logging / debugging."""
        totals = {
            "sent": {"bytes": 0, "messages": 0},
            "received": {"bytes": 0, "messages": 0},
        }
        for player in self._players.values():
            stats = player.stats
            totals["sent"]["bytes"] += stats["sent"]["bytes"]
            totals["sent"]["messages"] += stats["sent"]["messages"]
            totals["received"]["bytes"] += stats["received"]["bytes"]
            totals["received"]["messages"] += stats["received"]["messages"]
        return totals


    @property
    def timeout(self):
        """Amount of time allowed to elapse before blocking communications time-out.

        Returns
        -------
        timeout: number or `None`.
            The timeout in seconds, or `None` if there is no timeout.
        """
        return self._timeout


    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout


    @property
    def world_size(self):
        return self._world_size





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for user interaction."""

import numpy

from cicada.communicator.interface import Communicator

def secret_input(*, protocol, encoder, src, prompt=None, dtype=float):
    """Interactive prompt for a secret.

    Note
    ----
    Although this function only prompts one player for input, it is a
    collective operation that *must* be called by all players that are members
    of the communicator owned by `protocol`.

    Parameters
    ----------
    protocol: any protocol object, required
        Generates shares of the the secret value.
    src: int, required
        Rank of the player who will be prompted for a secret.
    prompt: str, optional
        Override the default interactive prompt.  See :func:`input`
        for usage.
    dtype: callable, optional
        Function for parsing user input into a secret-sharable value.  The
        function must take zero or one argument.  In the zero argument case, it
        should return a value which will be ignored.  In the one argument case,
        it must accept a string and return a scalar type that can be encoded
        and secret shared.  The builtin functions :func:`int` and :func:`float`
        are useful examples that provide this behavior.

    Returns
    -------
    share:
        A share of the secret input.  The type of the returned value will depend
        on the choice of protocol.
    """
    if protocol.communicator.rank == src:
        if prompt is None:
            prompt = f"Player {protocol.communicator.rank} secret: "
        secret = numpy.array(dtype(input(prompt)))
    else:
        secret = numpy.array(dtype())

    # Wait indefinitely for user input.
    with protocol.communicator.override(timeout=None):
        protocol.communicator.barrier()

    # Secret-share the input.
    return protocol.share(src=src, secret=encoder.encode(secret), shape=())



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for performing inter-process communication.

The classes in this submodule provide functions similar to those in MPI but
with additional API and semantics for reliability and error recovery.
"""

from .interface import Communicator
from .nng import NNGCommunicator
from .socket import SocketCommunicator




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for working with Shamir-shared secrets."""

import numbers

import numpy
import shamir

from cicada.communicator.interface import Communicator


class ShamirArrayShare(object):
    """Stores a local share of a Shamir-secret-shared array.

    Parameters
    ----------
    shape: :class:`tuple`, required
        The shape of the stored array.
    storage: :class:`object`, required
        Opaque storage for the secret-shared value.  This is created and
        manipulated by :class:`ShamirProtocol`, and is off-limits to all others!
    """
    def __init__(self, shape, storage):
        self._shape = shape
        self._storage = storage


    def __repr__(self):
        return f"cicada.shamir.ShamirArrayShare(shape={self._shape}, storage={self._storage})" # pragma: no cover


class ShamirProtocol(object):
    """Uses a communicator to create and manipulate Shamir-secret-shared values.

    Parameters
    ----------
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that this protocol will use for collective operations.
    """
    def __init__(self, communicator):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover
        self._communicator = communicator


    def _require_rank(self, rank, label):
        if not isinstance(rank, numbers.Integral):
            raise ValueError(f"Expected an integer for {label}, received {rank} instead.") # pragma: no cover
        if rank < 0 or rank >= self.communicator.world_size:
            raise ValueError(f"Expected {label} would be in the range [0, {self.communicator.world_size}), received {rank} instead.") # pragma: no cover
        return int(rank)


    def _require_rank_list(self, ranks, label):
        ranks = [self._require_rank(rank, label) for rank in ranks]
        if len(ranks) != len(set(ranks)):
            raise ValueError(f"Expected unique values for {label}, received {ranks} instead.") # pragma: no cover
        return ranks


    @property
    def communicator(self):
        """Returns the :class:`cicada.communicator.interface.Communicator` used by this protocol."""
        return self._communicator


    def reveal(self, share, src=None, dst=None):
        """Reconstruct a secret from its shares.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        share: :class:`ShamirArrayShare`, or :any:`None`, required
            A local share of the secret to be revealed.  This value is ignored
            if the local player isn't contributing a share.
        src: sequence of :class:`int`, optional
            List of players who will supply shares of the secret to be revealed.
            If :any:`None` (the default), all players will supply a share.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`,
            or :any:`None`.
        """

        # Identify who will be providing shares.
        if src is None:
            src = self.communicator.ranks

        # Identify who will be receiving shares.
        if dst is None:
            dst = self.communicator.ranks

        # Enforce preconditions.
        src = self._require_rank_list(src, "src")
        dst = self._require_rank_list(dst, "dst")

        if self.communicator.rank in src:
            if not isinstance(share, ShamirArrayShare):
                raise ValueError(f"share must be an instance of ShamirArrayShare, received {share} instead.") # pragma: no cover

        # Unpack data to exchange with the other players.
        value = (share._shape, share._storage) if self.communicator.rank in src else None

        # Send data to the other players.
        secret = None
        for recipient in dst:
            data = self.communicator.gatherv(src=src, value=value, dst=recipient)

            # If we're a recipient, recover the secret.
            if self.communicator.rank == recipient:
                shapes, shares = zip(*data)
                element_shares = numpy.array(shares).swapaxes(0, 1)
                secrets = [shamir.recover_secret(shares) for shares in element_shares]
                secret = numpy.array(secrets, dtype=object).reshape(shapes[0], order="C")

        return secret


    def share(self, *, src, secret, k, dst=None):
        """Distribute Shamir shares of a secret array.

        Note
        ----
        The input array must contain integers, and those integers
        must all be >= 0.

        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        src: integer, required
            The player providing the secret array to be shared.
        secret: :class:`numpy.ndarray` or :any:`None`, required
            The secret array to be shared, which must contain integers >= 0.
            This value is ignored for all players but `src`.
        k: integer, required
            Minimum number of shares required to reveal the secret.
            Must be <= the number of players that are members of
            :attr:`communicator`.
        dst: sequence of :class:`int`, optional.
            List of players who will receive shares of the secret.  If
            :any:`None` (the default), all players will receive a share.

        Returns
        -------
        share: :class:`ShamirArrayShare` or :any:`None`
            Local share of the shared secret, if this player is a member of
            `dst`, or :any:`None`.
        """
        # Identify who will be receiving shares.
        if dst is None:
            dst = self.communicator.ranks

        # Enforce preconditions.
        src = self._require_rank(src, "src")

        if self.communicator.rank == src:
            if not isinstance(secret, numpy.ndarray):
                raise ValueError(f"secret must be an instance of numpy.ndarray, received {secret} instead.") # pragma: no cover
            if not issubclass(secret.dtype.type, (numpy.integer, object)):
                raise ValueError(f"secret must contain integers, received {secret.dtype} instead.") # pragma: no cover
            if not numpy.all(secret >= 0):
                raise ValueError("secret must contain values >= 0.") # pragma: no cover

        k = int(k)
        if not (1 <= k and k <= len(dst)):
            raise ValueError(f"k must be an integer in the range [1, {len(dst)}], received {k} instead.") # pragma: no cover

        # Create a shamir share for each element in the secret.
        if self.communicator.rank == src:
            all_shares = []
            for element in secret.flat: # Returns elements in C-style (last index varies fastest) order.
                constant, shares = shamir.make_random_shares(k, len(dst))
                shares = numpy.array(shares)
                shares[:,1] -= constant
                shares[:,1] += element
                all_shares.append(shares)
            all_shares = numpy.array(all_shares)
            all_shares = all_shares.swapaxes(0, 1)
            values = [(secret.shape, shares) for shares in all_shares]
        else:
            values = None

        # Distribute shares to all players.
        shares = self.communicator.scatterv(src=src, values=values, dst=dst)

        # Package the shares.
        return shares if shares is None else ShamirArrayShare(shares[0], shares[1])




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import re

setup(
    name="cicada-mpc",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Communications",
        "Topic :: Scientific/Engineering",
        "Topic :: Security :: Cryptography",
        "Topic :: Utilities",
    ],
    description="Flexible toolkit for fault tolerant secure multiparty computation.",
    install_requires=[
        "netifaces",
        "numpy",
        "pynetstring",
        "pynng",
        "shamir",
    ],
    long_description="""Cicada is a flexible toolkit for working with fault-tolerant secure multiparty computation.  See the Cicada documentation at http://cicada-mpc.readthedocs.io, and the Cicada sources at http://github.com/cicada-mpc/cicada-mpc.""",
    maintainer="Timothy M. Shead",
    maintainer_email="tshead@sandia.gov",
    packages=find_packages(),
    project_urls={
        "Chat": "https://github.com/cicada-mpc/cicada-mpc/discussions",
        "Coverage": "https://coveralls.io/r/cicada-mpc/cicada-mpc",
        "Documentation": "https://cicada-mpc.readthedocs.io",
        "Issue Tracker": "https://github.com/cicada-mpc/cicada-mpc/issues",
        "Source": "https://github.com/cicada-mpc/cicada-mpc",
    },
    scripts=[
        "bin/cicada-exec",
    ],
    version=re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        open(
            "cicada/__init__.py",
            "r").read(),
        re.M).group(1),
)



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for binding to network interfaces."""

import contextlib
import logging
import socket

import netifaces

log = logging.getLogger()

def loopback_ip():
    """Return an IP address for localhost."""
    return "127.0.0.1"


def public_ip():
    """Return a public-facing IP address."""
    gateway, interface = netifaces.gateways()["default"][netifaces.AF_INET]
    return netifaces.ifaddresses(interface)[netifaces.AF_INET][0]["addr"]


def random_port(addr):
    """Return a randomly-chosen open port number for the given address."""
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind((addr, 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality to make logging from multiple processes easier."""

from cicada.communicator.interface import Communicator

class Logger(object):
    """Wrap a normal Python logger with a Cicada communicator to synchronize player output.

    Note
    ----
    Because :class:`Logger` communicates among players for synchronization, it
    can have a significant effect on performance, even when log messages are
    discarded by log levels or other filtering.

    Furthermore, :class:`Logger` should not be used in error recovery code, since it will
    fail attempting to communicate with players that are (presumably) dead.

    Parameters
    ----------
    logger: :class:`logging.Logger`, required.
        The Python logger to be used for output.
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that will be used to synchronize output among players.
    sync: boolean, optional
        Controls whether log output is synchronized among players.  In general,
        you should enable synchronization (the default) when multiple players
        are logging to the same console, for example during local debugging and
        regression tests.
    """
    def __init__(self, logger, communicator, sync=None):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover

        if sync is None:
            sync = True
        sync = bool(sync)

        self._logger = logger
        self._communicator = communicator
        self._sync = sync


    def _log(self, fn, src, *args, **kwargs):
        for i in range(self.communicator.world_size):
            if self.communicator.rank == i:
                if src is None or self.communicator.rank == src:
                    fn(*args, **kwargs)
            if self._sync:
                self._communicator.barrier()


    @property
    def communicator(self):
        """Returns the :class:`cicada.communicator.interface.Communicator` used by this logger."""
        return self._communicator


    def debug(self, *args, src=None, **kwargs):
        """Log a debug message, with optionally synchronized output among players.

        Note that this is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        """
        self._log(self._logger.debug, src, *args, **kwargs)


    def error(self, *args, src=None, **kwargs):
        """Log an error message, with optionally synchronized output among players.

        Note that this is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        """
        self._log(self._logger.error, src, *args, **kwargs)


    def info(self, *args, src=None, **kwargs):
        """Log an info message, with optionally synchronized output among players.

        Note that this is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        """
        self._log(self._logger.info, src, *args, **kwargs)


    def warning(self, *args, src=None, **kwargs):
        """Log a warning message, with optionally synchronized output among players.

        Note that this is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        """
        self._log(self._logger.warning, src, *args, **kwargs)

